==> thyi_generator-1.0.4.tar.gz: lib/src/thyi_generator.dart <==
import 'dart:async';
import 'dart:mirrors';

import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';
import 'package:thyi/thyi.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:code_builder/code_builder.dart';
import 'package:code_builder/code_builder.dart' as cb;
import 'package:dart_style/dart_style.dart';

class ThyiGenerator extends Generator {
  static Type typeGet = reflectClass(GET).reflectedType;
  static Type typePost = reflectClass(POST).reflectedType;
  static List<Type> typeHttpMethods = [typeGet, typePost];

  @override FutureOr<String> generate(LibraryReader library, BuildStep buildStep) {
    var inputId = buildStep.inputId;
    var file = _FileEx(library, inputId);
    if (!file.isApi()) {
      return null;
    }

    print("begin handle file: ${inputId.path}");
    return file.build(inputId);
  }
}

class _FileEx {
  static String PKG_THYI = "package:thyi/thyi.dart";
  List<_CleassElementEx> classes = [];
  List<String> pathSegments;
  bool hasInitialBuild = false;

  _FileEx(LibraryReader reader, AssetId assetId) {
    this.classes = reader.classes.map((clz) => _CleassElementEx(clz)).toList(growable: true);
    this.pathSegments = assetId.pathSegments;
  }

  bool isApi() {
    return "test" != this.pathSegments[0] && classes.any((c) => c.isApi());
  }

  void initialBuild() {
    if (hasInitialBuild) {
      return;
    }  

    this.classes.forEach((c) => c.initial());
  }

  String build(AssetId inputId) {
    initialBuild();
    final library = Library((b) => b
    ..directives.addAll(
      [Directive.import('dart:async')
      , Directive.import(this.pathSegments.last)
      , Directive.import(PKG_THYI)])
    ..body.addAll(this.classes.map((clz) => clz.build(inputId, b)).where((c) => c != null)));
    return DartFormatter().format('${library.accept(DartEmitter())}');
  }
}

class _CleassElementEx {
  static var PREFIX = "thyi";
  static var CLASS_PREFIX = "thyiImpl";
  List<_MethodElementEx> _methods = [];
  ClassElement element;
  String name;
  

  _CleassElementEx(this.element) {
    this._methods = element.methods.map((method) => _MethodElementEx(method)).toList(growable: true);
    this.name = "${element.name}_${CLASS_PREFIX}";
  }

  bool isApi() {
    return _methods.any((method) => method.isApi());
  }

  void initial() {
    this._methods.forEach((m) => m.initial());
  }

  Class build(AssetId inputId, cb.LibraryBuilder library) {
    var c = Class((b) => b
    ..name = this.name
    ..fields.addAll(this._buildFields())
    ..constructors.add(this._buildConstructor())
    ..extend = refer("${element.name}")
    ..methods.addAll(this._methods.map((method) => method.build(inputId, library)).where((method) => method != null))
    );
    return c;
  }

  List<Field> _buildFields() {
    List<Field> rst = [];
    rst.add(Field((b) {b
    ..name = "thyi"
    ..type = refer("Thyi");
    }));
    return rst;
  }

  Constructor _buildConstructor() {
    return Constructor((c) {c
    ..requiredParameters.add(Parameter((p) {p
      ..name = "thyi"
      ..toThis = true;
    }))
    ..body = Code('this.thyi = thyi;');
    });
  }
}

class _MethodElementEx {
  static String VAR_API_METHOD = "apiMethod";
  MethodElement element;
  _MethodParameters parameters;
  String returnType;
  Element returnTypeElement;

  _MethodElementEx(this.element);

  bool isApi() {
    return ThyiGenerator.typeHttpMethods.any((type) {
      return TypeChecker.fromRuntime(type).hasAnnotationOf(this.element, throwOnUnresolved: false);
    });
  }

  void initial() {
    this.parameters = _MethodParameters(element.parameters);
    var returnParameterizedType = this.element.returnType as ParameterizedType;
    returnTypeElement = returnParameterizedType.typeArguments[0].element;
    returnType = returnTypeElement.name;
  }

  Method build(AssetId inputId, cb.LibraryBuilder library) {
    if (!isApi()) {
      return null;
    }
    var returnType = element.returnType;
    if (!returnType.isDartAsyncFuture) {
      throw ArgumentError("for now can only support return Future");
    }

    this._buildReturnImport(inputId, library);

    List<Code> codes = [];
    codes.add(this._buildApiMethod());
    codes.add(this._buildReturn());
    
    var method = Method((b) => b
    ..name = element.displayName
    ..requiredParameters.addAll(parameters.buildRequiredParams())
    ..annotations.add(refer("override"))
    ..returns = refer(element.returnType.displayName, 'dart:async')
    ..body = Block.of(codes.where((code) => code != null).toList())
    );
    return method;
  }

  // ugly
  void _buildReturnImport(AssetId inputId, cb.LibraryBuilder library) {
    var package = inputId.package;
    var fullPath = returnTypeElement.library.source.toString();
    if (!fullPath.startsWith("dart:")) {
      fullPath = fullPath.substring(1);
      if (!fullPath.startsWith("${package}/")) {
        throw ArgumentError("doesn't impl import out package: ${fullPath}, current pkg: ${inputId.path}");
      }
      fullPath = fullPath.replaceAll("/lib/", "/");
      fullPath = "package:" + fullPath;
      library.directives.add(Directive.import(fullPath));
    }
  }

  Code _buildApiMethod() {
    var params = <Expression>[];
    final method = this._getHttpMethod();
    params.add(literal(method[0]));
    params.add(literal(method[1]));

    return refer("ApiMethod").newInstance(params, parameters.buildNamedParams()).assignFinal(VAR_API_METHOD).statement;
  }

  Code _buildReturn() {
    if (returnType == "String") {
      return refer("apiMethod.send(thyi)").returned.statement;
    } else {
      return refer("apiMethod.send(thyi).then((d) => ${returnType}.fromJson(jsonDecode(d)))").returned.statement;
    }
  }

  List<String> _getHttpMethod() {
    var meta = element.metadata.first;
    if (meta == null) {
      return ["GET", ""];
    }
    print("httpMethod: ${meta}");
    return [meta.constantValue.type.toString(), meta.constantValue.getField("path").toStringValue()];
  }

}

class _MethodParameters {
  List<ParameterElement> parameters;
  /// key: the annotation value, value: the param name
  Map<String, String> _headers;
  Map<String, String> _fields;
  Map<String, String> _quries;

  _MethodParameters(this.parameters) {
    this.parameters = this.parameters?? [];
    // _extraAnnotaion<FIELD>((meta) {
    //   meta.computeConstantValue();
    //   // print(meta);
    //   // print(meta.element);
    //   // print(meta.constantValue);
    //   // print(meta.computeConstantValue());
    //   // print(meta.constantValue);
    //   print(meta.constantValue.getField("key").toStringValue());
    //   return "";
    // });
    this._headers = _extraAnnotaion<HEADER>((meta) => meta.constantValue.getField("key").toStringValue());
    this._fields = _extraAnnotaion<FIELD>((meta) => meta.constantValue.getField("key").toStringValue());
    this._quries = _extraAnnotaion<QUERY>((meta) => meta.constantValue.getField("key").toStringValue());
  }

  List<Parameter> buildRequiredParams() {
    List<Parameter> rst = [];
    rst.addAll(parameters.map((pe) {
      return Parameter((b) => b
      ..name = pe.name
      ..type = refer(pe.type.name));
    }));
    return rst;
  }

  Map<String, Expression> buildNamedParams() {
    Map<String, Expression> namedArguments = {};
    _buildNamedParams(namedArguments, "headers", _headers);
    _buildNamedParams(namedArguments, "fields", _fields);
    _buildNamedParams(namedArguments, "queries", _quries);
    return namedArguments;
  }

  void _buildNamedParams<T>(Map<String, Expression> namedArguments, String name, Map<String, String> values) {
    if (values != null && values.isNotEmpty) {
      var genParam = values.entries.map((entry) => '"${entry.key}": ${entry.value}').join(",");
      print(genParam);
      namedArguments[name] = refer("{${genParam}}");
    }
  }

  Map<String, String> _extraAnnotaion<T>(String valueBuilder(ElementAnnotation meta)) {
    Map<String, String> rst = {};
    parameters.forEach((param) {
      if (param.metadata != null && param.metadata.isNotEmpty) {
        var meta = param.metadata.first;
        if (meta != null) {
          var metaName = meta.element.toString();
          if (metaName.startsWith("${T.toString()} ")) {
              meta.computeConstantValue();
              rst[valueBuilder(meta)] = param.name;
          }
        }
      }
    });
    return rst;
  }

}

==> thyi_generator-1.0.7.tar.gz: lib/src/thyi_generator.dart <==
import 'dart:async';
import 'dart:mirrors';

import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';
import 'package:thyi/thyi.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:code_builder/code_builder.dart';
import 'package:code_builder/code_builder.dart' as cb;
import 'package:dart_style/dart_style.dart';

class ThyiGenerator extends Generator {
  static Type typeGet = reflectClass(GET).reflectedType;
  static Type typePost = reflectClass(POST).reflectedType;
  static List<Type> typeHttpMethods = [typeGet, typePost];

  @override FutureOr<String> generate(LibraryReader library, BuildStep buildStep) {
    var inputId = buildStep.inputId;
    var file = _FileEx(library, inputId);
    if (!file.isApi()) {
      return null;
    }

    print("begin handle file: ${inputId.path}");
    return file.build(inputId);
  }
}

class _FileEx {
  static String PKG_THYI = "package:thyi/thyi.dart";
  List<_CleassElementEx> classes = [];
  List<String> pathSegments;
  bool hasInitialBuild = false;
  LibraryReader reader;

  _FileEx(this.reader, AssetId assetId) {
    this.classes = reader.classes.map((clz) => _CleassElementEx(clz, reader)).toList(growable: true);
    this.pathSegments = assetId.pathSegments;
  }

  bool isApi() {
    return "test" != this.pathSegments[0] && classes.any((c) => c.isApi());
  }

  void initialBuild() {
    if (hasInitialBuild) {
      return;
    }  

    this.classes.forEach((c) => c.initial());
  }

  List<Directive> _buildDirectives() {
    List<Directive> rst = [];
    rst.addAll([Directive.import('dart:async')
      , Directive.import('dart:convert')
      , Directive.import(this.pathSegments.last)
      , Directive.import(PKG_THYI)]);


    return rst;
  }

  String build(AssetId inputId) {
    print("aa");
    print(reader.element);
    print(reader.element.imports);
    print(reader.element.importedLibraries);
    print(reader.element.library);
    print(reader.element.library.imports);
    print(reader.element.library.importedLibraries);
    initialBuild();
    final library = Library((b) => b
    ..directives.addAll(_buildDirectives())
    ..body.addAll(this.classes.map((clz) => clz.build(inputId, b)).where((c) => c != null)));
    return DartFormatter().format('${library.accept(DartEmitter())}');
  }
}

class _CleassElementEx {
  static var PREFIX = "thyi";
  static var CLASS_PREFIX = "\$thyiImpl";
  List<_MethodElementEx> _methods = [];
  ClassElement element;
  LibraryReader libraryReader;
  String name;
  

  _CleassElementEx(this.element, this.libraryReader) {
    this._methods = element.methods.map((method) => _MethodElementEx(libraryReader, method)).toList(growable: true);
    this.name = "${element.name}${CLASS_PREFIX}";
  }

  bool isApi() {
    return _methods.any((method) => method.isApi());
  }

  void initial() {
    this._methods.forEach((m) => m.initial());
  }

  Class build(AssetId inputId, cb.LibraryBuilder library) {
    var c = Class((b) => b
    ..name = this.name
    ..fields.addAll(this._buildFields())
    ..constructors.add(this._buildConstructor())
    ..extend = refer("${element.name}")
    ..methods.addAll(this._methods.map((method) => method.build(inputId, library)).where((method) => method != null))
    );
    return c;
  }

  List<Field> _buildFields() {
    List<Field> rst = [];
    rst.add(Field((b) {b
    ..name = "thyi"
    ..type = refer("Thyi");
    }));
    return rst;
  }

  Constructor _buildConstructor() {
    return Constructor((c) {c
    ..requiredParameters.add(Parameter((p) {p
      ..name = "thyi"
      ..toThis = true;
    }))
    ..body = Code('this.thyi = thyi;');
    });
  }
}

class _MethodElementEx {
  static String VAR_API_METHOD = "apiMethod";
  MethodElement element;
  LibraryReader libraryReader;
  _MethodParameters parameters;
  String returnType;
  Element returnTypeElement;

  _MethodElementEx(this.libraryReader, this.element);

  bool isApi() {
    return ThyiGenerator.typeHttpMethods.any((type) {
      return TypeChecker.fromRuntime(type).hasAnnotationOf(this.element, throwOnUnresolved: false);
    });
  }

  void initial() {
    this.parameters = _MethodParameters(element.parameters);
    var returnParameterizedType = this.element.returnType as ParameterizedType;
    returnTypeElement = returnParameterizedType.typeArguments[0].element;
    returnType = returnTypeElement.name;
  }

  Method build(AssetId inputId, cb.LibraryBuilder library) {
    if (!isApi()) {
      return null;
    }
    var returnType = element.returnType;
    if (!returnType.isDartAsyncFuture) {
      throw ArgumentError("for now can only support return Future");
    }

    this._buildReturnImport(inputId, library);

    List<Code> codes = [];
    codes.add(this._buildApiMethod());
    codes.add(this._buildReturn());
    
    var method = Method((b) => b
    ..name = element.displayName
    ..requiredParameters.addAll(parameters.buildRequiredParams())
    ..optionalParameters.addAll(parameters.buildOptionParams())
    ..annotations.add(refer("override"))
    ..returns = refer(element.returnType.displayName, 'dart:async')
    ..body = Block.of(codes.where((code) => code != null).toList())
    );
    return method;
  }

  // ugly
  void _buildReturnImport(AssetId inputId, cb.LibraryBuilder library) {
    var package = inputId.package;
    var fullPath = returnTypeElement.library.source.toString();
    if (!fullPath.startsWith("dart:")) {
      // print("import: ${fullPath}");
      // print("to import: ${this.libraryReader.pathToElement(returnTypeElemnt)}");
      // fullPath = fullPath.substring(1);
      // if (!fullPath.startsWith("${package}/")) {
      //   throw ArgumentError("doesn't impl import out package: ${fullPath}, current pkg: ${inputId.path}");
      // }
      // fullPath = fullPath.replaceAll("/lib/", "/");
      // fullPath = "package:" + fullPath;
      library.directives.add(Directive.import(libraryReader.pathToElement(returnTypeElement).toString()));
    }
  }

  Code _buildApiMethod() {
    var params = <Expression>[];
    final method = this._getHttpMethod();
    params.add(literal(method[0]));
    params.add(literal(method[1]));

    return refer("ApiMethod").newInstance(params, parameters.buildNamedParams()).assignFinal(VAR_API_METHOD).statement;
  }

  Code _buildReturn() {
    if (returnType == "String") {
      return refer("apiMethod.send(thyi)").returned.statement;
    } else {
      return refer("apiMethod.send(thyi).then((d) => ${returnType}.fromJson(jsonDecode(d)))").returned.statement;
    }
  }

  List<String> _getHttpMethod() {
    var meta = element.metadata.first;
    if (meta == null) {
      return ["GET", ""];
    }
    print("httpMethod: ${meta}");
    return [meta.constantValue.type.toString(), meta.constantValue.getField("path").toStringValue()];
  }

}

class _MethodParameters {
  List<ParameterElement> parameters;
  /// key: the annotation value, value: the param name
  Map<String, String> _headers;
  Map<String, String> _fields;
  Map<String, String> _quries;

  _MethodParameters(this.parameters) {
    this.parameters = this.parameters?? [];
    this._headers = _extraAnnotaion<HEADER>((meta) => meta.constantValue.getField("key").toStringValue());
    this._fields = _extraAnnotaion<FIELD>((meta) => meta.constantValue.getField("key").toStringValue());
    this._quries = _extraAnnotaion<QUERY>((meta) => meta.constantValue.getField("key").toStringValue());
  }

  List<Parameter> buildOptionParams() {
    List<Parameter> rst = [];
    _buildParams(options: rst);
    return rst;
  }

  List<Parameter> buildRequiredParams() {
    List<Parameter> rst = [];
    _buildParams(requires: rst);
    return rst;
  }

  void _buildParams({List<Parameter> options, List<Parameter> requires}) {
    print("parameters' size: ${parameters.length}");
    parameters.forEach((pe) {
      print("map pe: ${pe}");
      var param = Parameter((b) {
        b.name = pe.name;
        b.type = refer(pe.type.name);
        b.named = pe.isNamed;
        if (pe.defaultValueCode != null) {
          b.defaultTo = refer(pe.defaultValueCode).code;
        }
      });
      if (pe.isOptional) {
        print("options add ${param}");
        options?.add(param);
      } else {
        print("requires add ${param}");
        requires?.add(param);
      }
    });
  }

  Map<String, Expression> buildNamedParams() {
    Map<String, Expression> namedArguments = {};
    _buildNamedParams(namedArguments, "headers", _headers);
    _buildNamedParams(namedArguments, "fields", _fields);
    _buildNamedParams(namedArguments, "queries", _quries);
    return namedArguments;
  }

  void _buildNamedParams<T>(Map<String, Expression> namedArguments, String name, Map<String, String> values) {
    if (values != null && values.isNotEmpty) {
      var genParam = values.entries.map((entry) => '"${entry.key}": ${entry.value}').join(",");
      print(genParam);
      namedArguments[name] = refer("{${genParam}}");
    }
  }

  Map<String, String> _extraAnnotaion<T>(String valueBuilder(ElementAnnotation meta)) {
    Map<String, String> rst = {};
    parameters.forEach((param) {
      if (param.metadata != null && param.metadata.isNotEmpty) {
        var meta = param.metadata.first;
        if (meta != null) {
          var metaName = meta.element.toString();
          if (metaName.startsWith("${T.toString()} ")) {
              meta.computeConstantValue();
              rst[valueBuilder(meta)] = param.name;
          }
        }
      }
    });
    return rst;
  }

}
